import numpy as np
from typing import List, NamedTuple, Callable, Optional, Union


# NamedTuple 三种取值：index name getattr()

class Dependency(NamedTuple):
    tensor: 'Tensor'
    grad_fn: Callable[[np.ndarray], np.ndarray]  # 张量依赖关系


IntArrayable = Union[tuple, list, np.ndarray]


def ensure_int_array(intarrayable: IntArrayable) -> np.ndarray:
    if isinstance(intarrayable, np.ndarray):
        return intarrayable
    else:
        return np.array(intarrayable, dtype=np.int)


Arrayable = Union[float, list, np.ndarray]


def ensure_array(arrayable: Arrayable) -> np.ndarray:
    if isinstance(arrayable, np.ndarray):
        return arrayable
    elif isinstance(arrayable, Tensor):
        return arrayable.data
    else:
        return np.array(arrayable, dtype=np.float64)


Tensorable = Union['Tensor', float, np.ndarray]


def ensure_tensor(tensorable: Tensorable) -> 'Tensor':
    if isinstance(tensorable, Tensor):
        return tensorable
    else:
        return Tensor(tensorable)


class Tensor:
    def __init__(self, data: Arrayable, requires_grad: bool = False, depends_on: List[Dependency] = None) -> None:
        self._data = ensure_array(data)
        self.requires_grad = requires_grad
        self.depends_on = depends_on or []
        self.shape = self.data.shape
        self.grad: Optional['Tensor'] = None
        if self.requires_grad:
            self.zero_grad()

    @property
    def data(self) -> np.ndarray:
        return self._data

    @data.setter
    def data(self, new_data: np.ndarray) -> None:
        self._data = new_data
        self.grad = None

    def zero_grad(self) -> None:
        self.grad = Tensor(np.zeros_like(self.data, dtype=np.float64))

    def __repr__(self) -> str:
        return f"Tensor({self.data},requires_grad={self.requires_grad})"

    def sum(self) -> 'Tensor':
        return tensor_sum(self)

    def __add__(self, other) -> 'Tensor':
        """
        get called if do t+ other
        :param other:
        :return:
        """
        return _add(self, ensure_tensor(other))

    def __radd__(self, other) -> 'Tensor':
        """
        get called if other + t
        :param other:
        :return:
        """
        return _add(ensure_tensor(other), self)

    def __iadd__(self, other) -> 'Tensor':
        """
        when we do t+=other
        :param other:
        :return:
        """
        self.data += ensure_tensor(other).data
        # invalidate the gradient
        return self

    def __isub__(self, other) -> 'Tensor':
        self.data -= ensure_tensor(other).data
        # invalidate the gradient
        return self

    def __imul__(self, other) -> 'Tensor':
        self.data *= ensure_tensor(other).data
        # invalidate the gradient
        return self

    def __mul__(self, other) -> 'Tensor':
        return _mul(self, ensure_tensor(other))

    def __rmul__(self, other) -> 'Tensor':
        return _mul(ensure_tensor(other), self)

    def __matmul__(self, other) -> 'Tensor':
        return _matmul(self, ensure_tensor(other))

    def __neg__(self) -> 'Tensor':
        return _neg(self)

    def __sub__(self, other) -> 'Tensor':
        return _sub(self, ensure_tensor(other))

    def __rsub__(self, other) -> 'Tensor':
        return _sub(ensure_tensor(other), self)

    def __getitem__(self, item) -> 'Tensor':
        return _slice(self, item)

    def backward(self, grad: 'Tensor' = None) -> None:
        assert self.requires_grad, "called backward on non-requires-grad tensor"

        if grad is None:
            if self.shape == ():
                grad = Tensor(1.0)
            else:
                raise RuntimeError("grad must be specified for non-0-Tensor")

        self.grad.data += grad.data
        for dependency in self.depends_on:
            backward_grad = dependency.grad_fn(grad.data)
            dependency.tensor.backward(Tensor(backward_grad))

    def transpose(self, intarrayable: IntArrayable) -> 'Tensor':
        data = self.data.transpose(ensure_int_array(intarrayable))
        requires_grad = self.requires_grad
        if requires_grad:
            def grad_fn(grad: np.ndarray) -> np.ndarray:
                """
                grad is necessarily a 0-tenosr, so each input element contributes that much
                :param grad:
                :return:
                """
                array = ensure_int_array(intarrayable).squeeze()
                index = [x for x in range(len(array))]
                index_array = dict(zip(array, index))
                index_array = list(sorted(index_array.items(), key=lambda item: item[0]))
                index_array = [y[1] for x, y in enumerate(index_array)]
                return grad.transpose(ensure_int_array(index_array))

            depends_on = [Dependency(self, grad_fn)]

        else:
            depends_on = []

        return Tensor(data, requires_grad, depends_on)



def tensor_sum(t: Tensor) -> Tensor:
    """
    tankes a tensor and returns the 0-tensor
    that's the sum of all its elemetns;
    :param t:
    :return:
    """
    data = t.data.sum()
    requires_grad = t.requires_grad
    if requires_grad:
        def grad_fn(grad: np.ndarray) -> np.ndarray:
            """
            grad is necessarily a 0-tenosr, so each input element contributes that much
            :param grad:
            :return:
            """
            return grad * np.ones_like(t.data)

        depends_on = [Dependency(t, grad_fn)]

    else:
        depends_on = []

    return Tensor(data, requires_grad, depends_on)


def _add(t1: Tensor, t2: Tensor) -> Tensor:
    data = t1.data + t2.data
    requires_grad = t1.requires_grad or t2.requires_grad
    depends_on: List[Dependency] = []

    if t1.requires_grad:
        def grad_fn1(grad: np.ndarray) -> np.ndarray:
            # handel broadcasting properly

            # sum out added dims
            # (2,3)+(3,)
            ndims_added = grad.ndim - t1.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(axis=0)

            # sum across braodcasted(but non-added dims)
            # (2,3) + (1,3)
            for i, dim in enumerate(t1.shape):
                if dim == 1:
                    grad = grad.sum(axis=i, keepdims=True)
            return grad

        depends_on.append(Dependency(t1, grad_fn1))

    if t2.requires_grad:
        def grad_fn2(grad: np.ndarray) -> np.ndarray:
            ndims_added = grad.ndim - t2.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(axis=0)

            for i, dim in enumerate(t2.shape):
                if dim == 1:
                    grad = grad.sum(axis=i, keepdims=True)

            return grad

        depends_on.append(Dependency(t2, grad_fn2))

    return Tensor(data, requires_grad, depends_on)


def _mul(t1: Tensor, t2: Tensor) -> Tensor:
    """
    y = a * b
    have dL/dy
    dL/da = dL/dy * b
    :param t1:
    :param t2:
    :return:
    """
    data = t1.data * t2.data
    requires_grad = t1.requires_grad or t2.requires_grad
    depends_on: List[Dependency] = []

    if t1.requires_grad:
        def grad_fn1(grad: np.ndarray) -> np.ndarray:
            # handel broadcasting properly
            grad = grad * t2.data
            # sum out added dims
            # (2,3)+(3,)
            ndims_added = grad.ndim - t1.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(axis=0)

            # sum across braodcasted(but non-added dims)
            # (2,3) + (1,3)
            for i, dim in enumerate(t1.shape):
                if dim == 1:
                    grad = grad.sum(axis=i, keepdims=True)
            return grad

        depends_on.append(Dependency(t1, grad_fn1))

    if t2.requires_grad:
        def grad_fn2(grad: np.ndarray) -> np.ndarray:
            grad = grad * t1.data
            ndims_added = grad.ndim - t2.data.ndim
            for _ in range(ndims_added):
                grad = grad.sum(axis=0)

            for i, dim in enumerate(t2.shape):
                if dim == 1:
                    grad = grad.sum(axis=i, keepdims=True)

            return grad

        depends_on.append(Dependency(t2, grad_fn2))

    return Tensor(data, requires_grad, depends_on)


def _neg(t: Tensor) -> Tensor:
    data = -t.data
    requires_grad = t.requires_grad
    if requires_grad:
        depends_on = [Dependency(t, lambda x: -x)]

    else:
        depends_on = []

    return Tensor(data, requires_grad, depends_on)


def _sub(t1: Tensor, t2: Tensor) -> Tensor:
    return t1 + -t2


def _matmul(t1: Tensor, t2: Tensor) -> Tensor:
    """
        if t1 =(n1,m1) and t2 is (m1,m2) then t1 @ t2 = (n1,m2)
        so grad3  is (n1,m2)

        if t3 = t1 @ t2
        and grad3 if the gradient of some function wrt t3 ,
        then
        grad1 = grad3 @t2.T (n1,m2) * (m2,m1)
        grad2 = t1.T @ grad3 (m1,n1) * (n1.m2)


        """
    data = t1.data @ t2.data
    requires_grad = t1.requires_grad or t2.requires_grad
    depends_on: List[Dependency] = []

    if t1.requires_grad:
        def grad_fn1(grad: np.ndarray) -> np.ndarray:
            # handel broadcasting properly
            return grad @ t2.data.T

        depends_on.append(Dependency(t1, grad_fn1))

    if t2.requires_grad:
        def grad_fn2(grad: np.ndarray) -> np.ndarray:
            return t1.data.T @ grad

        depends_on.append(Dependency(t2, grad_fn2))

    return Tensor(data, requires_grad, depends_on)


def _slice(t: Tensor, idx) -> Tensor:
    """
    :param t:
    :param idx:
    :return:
    """
    data = t.data[idx]
    requires_grad = t.requires_grad

    if requires_grad:
        def grad_fn(grad: np.ndarray) -> np.ndarray:
            bigger_grad = np.zeros_like(data)
            bigger_grad[idx] = grad
            return bigger_grad

        depends_on = Dependency(t, grad_fn)

    else:
        depends_on = []

    return Tensor(data, requires_grad, depends_on)



from autograd import Tensor, Parameter, Module
from autograd.Layers import Linear
import numpy as np


class SimpleAttention(Module):
    def __init__(self, in_size: int = 100) -> None:
        self.size = in_size

    def __call__(self, q, k, v):
        return self.predict(q, k, v)

    def predict(self, q: Tensor, k: Tensor, v: Tensor) -> Tensor:
        # q,k,v (batch_size,,length,size)
        # q*k -> (batch_size,head,length,length)
        # q*k*v->(batch_size,head,length,size)
        attn = q @ k.transpose((-1,-2))
        output = attn @ v
        return output



import unittest
from autograd import Tensor

import numpy as np


class TestTensorTranspose(unittest.TestCase):
    def test_simple_add(self):
        t1 = Tensor(np.random.randn(3, 4, 5), requires_grad=True)
        t2 = t1.transpose((1, 2, 0))
        assert t2.data.shape == (4, 5, 3)
        t2.backward(Tensor(np.ones((4, 5, 3))))
        assert t1.grad.data.shape == (3, 4, 5)



import unittest
from autograd import Tensor


class TestTensorSum(unittest.TestCase):
    def test_simple_add(self):
        t1 = Tensor([1, 2, 3], requires_grad=True)
        t2 = Tensor([4, 5, 6], requires_grad=True)
        t3 = t1 + t2
        assert t3.data.tolist() == [5, 7, 9]
        t3.backward(Tensor([-1, -2, -3]))

        assert t1.grad.data.tolist() == [-1, -2, -3]
        assert t2.grad.data.tolist() == [-1, -2, -3]

        t1 += 0.1
        assert t1.grad is None
        assert t1.data.tolist() == [1.1, 2.1, 3.1]

    def test_broadcast_add(self):
        """
        eg: t1.shape==(10,5)  t2.shape=(5,) => t1 + t2  ,viewed as(1,5)
            t2=[1,2,3,4,5] => viewed v2 as[[1,2,3,4,5]]

        eg:
            t1 as (10,5)
            t2 as (1,5) ias [[1,2,3,4,,5]]
        :return:
        """
        t1 = Tensor([[1, 2, 3], [4, 5, 6]], requires_grad=True)
        t2 = Tensor([7, 8, 9], requires_grad=True)

        t3 = t1 + t2  # shape(2,3)

        assert t3.data.tolist() == [[8, 10, 12], [11, 13, 15]]

        t3.backward(Tensor([[1, 1, 1], [1, 1, 1]]))

        assert t1.grad.data.tolist() == [[1, 1, 1], [1, 1, 1]]
        assert t2.grad.data.tolist() == [2, 2, 2]

    def test_broadcast_add2(self):
        t1 = Tensor([[1, 2, 3], [4, 5, 6]], requires_grad=True)
        t2 = Tensor([[7, 8, 9]], requires_grad=True)

        t3 = t1 + t2  # shape(2,3)
        assert t3.data.tolist() == [[8, 10, 12], [11, 13, 15]]

        t3.backward(Tensor([[1, 1, 1], [1, 1, 1]]))

        assert t1.grad.data.tolist() == [[1, 1, 1], [1, 1, 1]]
        assert t2.grad.data.tolist() == [[2, 2, 2]]



from autograd import Tensor, Parameter, Module


class Linear(Module):
    def __init__(self, in_size: int = 100, hidden_size: int = 50) -> None:
        self.w1 = Parameter(in_size, hidden_size)
        self.b1 = Parameter(hidden_size)

    def __call__(self, inputs: Tensor) -> Tensor:
        return self.predict(inputs)

    def predict(self, inputs: Tensor) -> Tensor:
        # inputs (batch_size,10)
        x1 = inputs @ self.w1 + self.b1  # (batch_size,num_hidden)

        return x1



import unittest
from autograd import Tensor


class TestTensorSum(unittest.TestCase):
    def test_simple_sum(self):
        t1 = Tensor([1, 2, 3], requires_grad=True)
        t2 = t1.sum()
        t2.backward()

        assert t1.grad.data.tolist() == [1, 1, 1]

    def test_sum_with_grad(self):
        t1 = Tensor([1, 2, 3], requires_grad=True)
        t2 = t1.sum()
        t2.backward(Tensor(3))

        assert t1.grad.data.tolist() == [3, 3, 3]



import unittest
from autograd import Tensor
import numpy as np


class TestTensorMatMul(unittest.TestCase):
    def test_simple_mul(self):
        t1 = Tensor([[1, 2], [3, 4], [5, 6]], requires_grad=True)
        t2 = Tensor([[10], [20]], requires_grad=True)
        t3 = t1 @ t2
        assert t3.data.tolist() == [[50], [110], [170]]
        grad = Tensor([[-1], [-2], [-3]])
        t3.backward(grad)

        assert np.all(t1.grad.data == grad.data @ t2.data.T)
        assert np.all(t2.grad.data == t1.data.T @ grad.data )



import unittest
from autograd import Tensor
import numpy as np


class TestTensorMul(unittest.TestCase):
    def test_simple_mul(self):
        t1 = Tensor([1, 2, 3], requires_grad=True)
        t2 = Tensor([4, 5, 6], requires_grad=True)
        t3 = t1 * t2
        assert t3.data.tolist() == [4, 10, 18]
        t3.backward(Tensor([-1, -2, -3]))

        assert t1.grad.data.tolist() == [-4, -10, -18]
        assert t2.grad.data.tolist() == [-1, -4, -9]

        t1 *= 0.1
        assert t1.grad is None
        assert np.allclose(t1.data, [0.1, 0.2, 0.3])
        # assert t1.data.tolist() == [0.1, 0.2, 0.3]

    def test_broadcast_mul(self):
        t1 = Tensor([[1, 2, 3], [4, 5, 6]], requires_grad=True)
        t2 = Tensor([7, 8, 9], requires_grad=True)

        t3 = t1 * t2  # shape(2,3)
        assert t3.data.tolist() == [[7, 16, 27], [28, 40, 54]]
        t3.backward(Tensor([[1, 1, 1], [1, 1, 1]]))

        assert t1.grad.data.tolist() == [[7, 8, 9], [7, 8, 9]]
        assert t2.grad.data.tolist() == [5, 7, 9]

    def test_broadcast_mul2(self):
        t1 = Tensor([[1, 2, 3], [4, 5, 6]], requires_grad=True)
        t2 = Tensor([[7, 8, 9]], requires_grad=True)

        t3 = t1 * t2  # shape(2,3)
        assert t3.data.tolist() == [[7, 16, 27], [28, 40, 54]]
        t3.backward(Tensor([[1, 1, 1], [1, 1, 1]]))

        assert t1.grad.data.tolist() == [[7, 8, 9], [7, 8, 9]]
        assert t2.grad.data.tolist() == [[5, 7, 9]]



from autograd.tensor import Tensor
import numpy as np

"""
to quick new a tensor with requires grad 
also,
optimize for module
"""


class Parameter(Tensor):
    def __init__(self, *shape):
        data = np.random.randn(*shape)
        super(Parameter, self).__init__(data, requires_grad=True)



"""
Optimizers go here
"""

from autograd import Module


class SGD:
    def __init__(self, lr: float = 0.01) -> None:
        self.lr = lr

    def step(self, module: Module) -> None:
        for parameter in module.parameters():
            parameter -= parameter.grad * self.lr



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="Torch-Yottaxx",
    version="0.1.3",
    packages=find_packages(),
    author="Yotta",
    author_email="xiaoz987@gmail.com",
    description="a simple autograd library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Yottaxx/YottaTorch",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy',
        'pytest'
    ]
)



from autograd.tensor import Tensor, Dependency
import numpy as np


def tanh(tensor: Tensor) -> Tensor:
    data = np.tanh(tensor.data)
    requires_grad = tensor.requires_grad
    if requires_grad:
        def grad_fn(grad: np.ndarray) -> np.ndarray:
            return grad * (1 - data * data)

        depends_on = [Dependency(tensor, grad_fn)]
    else:
        depends_on = []

    return Tensor(data, requires_grad, depends_on)

